import json
import uuid
from typing import Any, Optional
from unittest.mock import MagicMock

import pytest
from bunch import Bunch
from hamcrest import *
from pytest_tools import pipe, scenariostep

from melange import DriverManager
from melange.drivers import configure_exchange
from melange.drivers.interfaces import MessagingDriver, Message
from melange.event_serializer import MessageSerializer
from melange.exchange_listener import ExchangeListener, listener
from melange.exchange_message_consumer import ExchangeMessageConsumer
from melange.exchange_message_publisher import ExchangeMessagePublisher, SQSPublisher


class TestEventSerializer(MessageSerializer):
    def manifest(self, data: Any) -> str:
        return "BananaEvent"

    def serialize(self, data: Any) -> str:
        if isinstance(data, BananaEvent):
            return json.dumps({"value": data.somevalue})
        return "apple"

    def deserialize(self, serialized_data: str, manifest: Optional[str] = None) -> Any:
        if serialized_data != "apple":
            data = json.loads(serialized_data)
            return BananaEvent(somevalue=data["value"])


class BananaEvent:
    def __init__(self, somevalue):
        self.somevalue = somevalue


class TestMessageConsumer:
    def test_consume_on_empty_queue(self, testcontext):
        event_type_name = "an_event_type_event"

        pipe(
            testcontext,
            given_an_empty_queue_to_listen(),
            given_a_subscriber(listens_to=event_type_name),
            when_consuming_the_next_event_in_queue(),
            then_the_subscriber_has_not_processed_any_event(),
        )

    def test_consume_on_queue_with_message(self):
        event_type_name = "an_event_type_event"

        self._Scenario().given_a_queue_to_listen_with_an_event(
            event_of_type=event_type_name
        ).given_a_subscriber(
            listens_to=event_type_name
        ).when_consuming_the_next_event_in_queue().then_the_subscriber_has_processed_the_event().the_message_has_been_acknowledged()

    def test_consume_on_queue_with_message_but_subscriber_listens_to_different_events(
        self,
    ):
        event_type_name = "an_event_type_event"
        event_type_name_2 = "an_event_type_event_2"

        self._Scenario().given_a_queue_to_listen_with_an_event(
            event_of_type=event_type_name
        ).given_a_subscriber(
            listens_to=event_type_name_2
        ).when_consuming_the_next_event_in_queue().then_the_subscriber_has_not_processed_any_event().the_message_has_been_acknowledged()

    def test_unsubscribe(self):
        event_type_name = "an_event_type_event"

        self._Scenario().given_a_queue_to_listen_with_an_event(
            event_of_type=event_type_name
        ).given_a_subscriber(
            listens_to=event_type_name
        ).when_unsubscribing_the_subscriber().when_consuming_the_next_event_in_queue().then_the_subscriber_has_not_processed_any_event().the_message_has_been_acknowledged()

    class _Scenario:
        def __init__(self):
            self.message_consumer = None
            self.subscriber = None
            self.event = None
            self.messages = None
            self.driver = MagicMock(spec=MessagingDriver)
            DriverManager().use_driver(driver=self.driver)

        def given_a_queue_to_listen_with_an_event(self, event_of_type=None):
            self.messages = [self._create_message(i) for i in range(1)]
            queue = self._create_queue(self.messages)
            self.driver.declare_queue.return_value = (queue, "")

            self.event = Message(uuid.uuid4(), {"event_type_name": event_of_type}, None)
            self.driver.retrieve_messages.return_value = [self.event]

            event_queue_name = "a_queue_name"
            topic_to_subscribe = "a_topic_name"
            self.message_consumer = ExchangeMessageConsumer(
                event_queue_name, TestEventSerializer(), topic_to_subscribe
            )

            return self

        def given_a_subscriber(self, listens_to=None):
            self.subscriber = MagicMock(spec=ExchangeListener)
            self.subscriber.listens_to.return_value = listens_to

            self.message_consumer.subscribe(self.subscriber)

            return self

        def when_consuming_the_next_event_in_queue(self):
            self.message_consumer.consume_event()

            return self

        def then_the_subscriber_has_processed_the_event(self):
            self.subscriber.process_event.assert_called()
            return self

        def the_message_has_been_acknowledged(self):
            self.driver.acknowledge.assert_called()

            return self

        def given_an_empty_queue_to_listen(self):
            queue = self._create_queue()
            self.driver.declare_queue = MagicMock(return_value=(queue, ""))
            self.driver.declare_topic = MagicMock()

            event_queue_name = "a_queue_name"
            topic_to_subscribe = "a_topic_name"
            self.message_consumer = ExchangeMessageConsumer(
                event_queue_name, TestEventSerializer(), topic_to_subscribe
            )

            return self

        def then_the_subscriber_has_not_processed_any_event(self):
            self.subscriber.process.assert_not_called()
            return self

        def _create_queue(self, messages=[]):
            queue = MagicMock()

            self.driver.retrieve_messages.return_value = messages

            return queue

        def _create_message(self, i):
            message = MagicMock()
            message.body = json.dumps({"event_type_name": "an_event_name"})
            return message

        def when_unsubscribing_the_subscriber(self):
            self.message_consumer.unsubscribe(self.subscriber)
            return self


class TestMessageConsumerWithAWS:
    @pytest.fixture(autouse=True)
    def setup(self, request):
        self.exchange_consumer = None
        configure_exchange("aws")

        def teardown():
            driver = DriverManager().get_driver()
            if self.exchange_consumer:
                if self.exchange_consumer._event_queue:
                    driver.delete_queue(self.exchange_consumer._event_queue)
                if self.exchange_consumer._dead_letter_queue:
                    driver.delete_queue(self.exchange_consumer._dead_letter_queue)
                if self.exchange_consumer._topics:
                    for topic in self.exchange_consumer._topics:
                        driver.delete_topic(topic)

        request.addfinalizer(teardown)

    def test_consume_event_from_sqs(self):
        topic_name = self._get_topic_name()
        serializer = TestEventSerializer()
        self.exchange_consumer = ExchangeMessageConsumer(
            self._get_queue_name(), serializer, topic_name
        )
        exchange_publisher = ExchangeMessagePublisher(
            topic=topic_name, message_serializer=serializer
        )

        class TestListener(ExchangeListener):
            def __init__(self):
                super().__init__()
                self.listened_event = None

            @listener
            def on_event(self, event: BananaEvent):
                self.listened_event = event

        test_listener = TestListener()
        self.exchange_consumer.subscribe(test_listener)
        exchange_publisher.publish(BananaEvent("somevalue"))

        self.exchange_consumer.consume_event()
        assert_that(test_listener.listened_event, is_(BananaEvent))
        assert_that(test_listener.listened_event.somevalue, is_("somevalue"))

    def test_consume_event_coming_from_the_sqs_publisher(self):
        topic_name = self._get_queue_name()
        serializer = TestEventSerializer()
        queue_name = self._get_queue_name()
        self.exchange_consumer = ExchangeMessageConsumer(
            queue_name, serializer, topic_name
        )
        queue_publisher = SQSPublisher(
            queue_name=queue_name, message_serializer=serializer
        )

        class TestListener(ExchangeListener):
            def __init__(self):
                super().__init__()
                self.listened_event = None

            @listener
            def on_event(self, event: BananaEvent):
                self.listened_event = event

        test_listener = TestListener()
        self.exchange_consumer.subscribe(test_listener)
        queue_publisher.publish(BananaEvent("somevalue"))

        self.exchange_consumer.consume_event()
        assert_that(test_listener.listened_event, is_(BananaEvent))
        assert_that(test_listener.listened_event.somevalue, is_("somevalue"))

    def test_consume_event_with_listeners_that_listen_multiple_events(self):
        topic_name = self._get_topic_name()
        self.exchange_consumer = ExchangeMessageConsumer(
            self._get_queue_name(), TestEventSerializer(), topic_name
        )
        exchange_publisher = ExchangeMessagePublisher(
            topic=topic_name, message_serializer=TestEventSerializer()
        )

        self.listened_events = set()

        class TestListener(ExchangeListener):
            def __init__(self):
                super().__init__(None)

            def process(x, event, **kwargs):
                self.listened_events.add(event["value"])

            def listens_to(x):
                return ["TestEvent", "TestEvent2", "TestEvent3"]

        self.exchange_consumer.subscribe(TestListener())
        exchange_publisher.publish({"value": "somevalue"})
        exchange_publisher.publish(
            {"value": "somevalue2", "event_type_name": "TestEvent2"}
        )
        exchange_publisher.publish({"value": "somevalue3"})
        exchange_publisher.publish({"value": "somevalue4"})

        self.exchange_consumer.consume_event()
        self.exchange_consumer.consume_event()
        self.exchange_consumer.consume_event()
        self.exchange_consumer.consume_event()

        assert len(self.listened_events) == 3
        assert "somevalue" in self.listened_events
        assert "somevalue2" in self.listened_events
        assert "somevalue4" in self.listened_events
        assert "somevalue3" not in self.listened_events

    def _get_queue_name(self):
        return "test_queue_{}".format(uuid.uuid4())

    def _get_topic_name(self):
        return "test_queue_{}".format(uuid.uuid4())


@pytest.fixture
def testcontext():
    driver = MagicMock(spec=MessagingDriver)
    DriverManager().use_driver(driver=driver)
    return Bunch(
        message_consumer=None, subscriber=None, event=None, messages=None, driver=driver
    )


def given_a_queue_to_listen_with_an_event(self, event_of_type=None):
    self.messages = [self._create_message(i) for i in range(1)]
    queue = self._create_queue(self.messages)
    self.driver.declare_queue.return_value = (queue, "")

    self.event = Message(uuid.uuid4(), {"event_type_name": event_of_type}, None)
    self.driver.retrieve_messages.return_value = [self.event]

    event_queue_name = "a_queue_name"
    topic_to_subscribe = "a_topic_name"
    self.message_consumer = ExchangeMessageConsumer(
        event_queue_name, TestEventSerializer(), topic_to_subscribe
    )

    return self


@scenariostep
def given_a_subscriber(ctx, listens_to=None):
    ctx.subscriber = MagicMock(spec=ExchangeListener)
    ctx.subscriber.listens_to.return_value = listens_to

    ctx.message_consumer.subscribe(ctx.subscriber)

    return ctx


@scenariostep
def when_consuming_the_next_event_in_queue(ctx):
    ctx.message_consumer.consume_event()

    return ctx


def then_the_subscriber_has_processed_the_event(self):
    self.subscriber.process_event.assert_called()
    return self


def the_message_has_been_acknowledged(self):
    self.driver.acknowledge.assert_called()

    return self


@scenariostep
def given_an_empty_queue_to_listen(ctx):
    driver = ctx.driver
    queue = _create_queue(driver)
    driver.declare_queue = MagicMock(return_value=(queue, ""))
    driver.declare_topic = MagicMock()

    event_queue_name = "a_queue_name"
    topic_to_subscribe = "a_topic_name"
    ctx.message_consumer = ExchangeMessageConsumer(event_queue_name, topic_to_subscribe)

    return ctx


@scenariostep
def then_the_subscriber_has_not_processed_any_event(ctx):
    ctx.subscriber.process.assert_not_called()
    return ctx


def _create_queue(driver, messages=None):
    queue = MagicMock()

    driver.retrieve_messages.return_value = messages or []

    return queue


def _create_message(self, i):
    message = MagicMock()
    message.body = json.dumps({"event_type_name": "an_event_name"})
    return message


def when_unsubscribing_the_subscriber(self):
    self.message_consumer.unsubscribe(self.subscriber)


@pytest.mark.skip()
class TestMessageConsumerRabbitMQ:
    def setup_method(self, m):
        self.exchange_consumer = None
        DriverManager().use_driver(driver_name="rabbitMQ", host="localhost")
        self.driver = DriverManager().get_driver()

    def teardown_method(self):
        if self.exchange_consumer._topics:
            for topic in self.exchange_consumer._topics:
                self.driver.delete_topic(topic)

        if self.exchange_consumer._event_queue:
            self.driver.delete_queue(self.exchange_consumer._event_queue)

        if self.exchange_consumer._dead_letter_queue:
            self.driver.delete_queue(self.exchange_consumer._dead_letter_queue)

    def test_consume_event_from_queue(self):
        topic_name = self._get_topic_name()
        self.exchange_consumer = ExchangeMessageConsumer(
            self._get_queue_name(), topic_name
        )
        exchange_publisher = ExchangeMessagePublisher(topic=topic_name)

        self.listened_event = None

        class TestListener(ExchangeListener):
            def process(x, event, **kwargs):
                self.listened_event = event

            def listens_to(self):
                return ["TestEvent"]

        self.exchange_consumer.subscribe(TestListener())
        exchange_publisher.publish({"value": "somevalue"}, manifest="TestEvent")

        self.exchange_consumer.consume_event()
        assert self.listened_event["value"] == "somevalue"
        assert self.listened_event["event_type_name"] == "TestEvent"

    def test_consume_event_with_listeners_that_listen_multiple_events(self):
        topic_name = self._get_topic_name()
        self.exchange_consumer = ExchangeMessageConsumer(
            self._get_queue_name(), topic_name
        )
        exchange_publisher = ExchangeMessagePublisher(topic=topic_name)

        self.listened_events = set()

        class TestListener(ExchangeListener):
            def process(x, event, **kwargs):
                self.listened_events.add(event["value"])

            def listens_to(x):
                return ["TestEvent", "TestEvent2", "TestEvent3"]

        self.exchange_consumer.subscribe(TestListener())
        exchange_publisher.publish({"value": "somevalue"}, manifest="TestEvent")
        exchange_publisher.publish(
            {"value": "somevalue2", "event_type_name": "TestEvent2"}
        )
        exchange_publisher.publish({"value": "somevalue3"}, manifest="TestEvent4")
        exchange_publisher.publish({"value": "somevalue4"}, manifest="TestEvent3")

        self.exchange_consumer.consume_event()
        self.exchange_consumer.consume_event()
        self.exchange_consumer.consume_event()
        self.exchange_consumer.consume_event()

        assert len(self.listened_events) == 3
        assert "somevalue" in self.listened_events
        assert "somevalue2" in self.listened_events
        assert "somevalue4" in self.listened_events
        assert "somevalue3" not in self.listened_events

    def _get_queue_name(self):
        return "test_queue_{}".format(uuid.uuid4())

    def _get_topic_name(self):
        return "test_topic_{}".format(uuid.uuid4())



import logging
import uuid

from melange.drivers.driver_manager import DriverManager
from melange.event_serializer import MessageSerializer

logger = logging.getLogger(__name__)


class ExchangeMessagePublisher:
    def __init__(self, message_serializer: MessageSerializer, topic, driver=None):
        self._driver = driver or DriverManager().get_driver()
        self.message_serializer = message_serializer
        self._topic_name = topic

    def init(self):
        self._topic = self._driver.declare_topic(self._topic_name)

    def publish(self, data, extra_attributes=None):
        content = self.message_serializer.serialize(data)
        manifest = self.message_serializer.manifest(data)

        self.init()
        self._driver.publish(
            content,
            self._topic,
            event_type_name=manifest,
            extra_attributes=extra_attributes,
        )

        return True


class SQSPublisher:
    def __init__(
        self,
        queue_name,
        message_serializer: MessageSerializer,
        dlq_name=None,
        driver=None,
        **kwargs
    ):
        self._driver = driver or DriverManager().get_driver()
        self.message_serializer = message_serializer
        self._event_queue, self._dead_letter_queue = self._driver.declare_queue(
            queue_name, dead_letter_queue_name=dlq_name
        )
        self.default_message_group_id = kwargs.get("message_group_id")
        self.is_fifo = self._event_queue.attributes.get("FifoQueue") == "true"

    def publish(self, data, **kwargs):
        content = self.message_serializer.serialize(data)
        manifest = self.message_serializer.manifest(data)

        message_group_id = kwargs.get("message_group_id", self.default_message_group_id)
        message_deduplication_id = (
            None
            if not self.is_fifo
            else kwargs.get("message_deduplication_id", str(uuid.uuid4()))
        )

        self._driver.queue_publish(
            content,
            self._event_queue,
            event_type_name=manifest,
            message_group_id=message_group_id,
            message_deduplication_id=message_deduplication_id,
        )



import os

CACHE_REDIS_HOST = os.environ.get("CACHE_REDIS_HOST")
CACHE_REDIS_PORT = os.environ.get("CACHE_REDIS_PORT")
CACHE_REDIS_DB = os.environ.get("CACHE_REDIS_DB")
CACHE_REDIS_PASSWORD = os.environ.get("CACHE_REDIS_PASSWORD")
CACHE_NAMESPACE = os.environ.get("CACHE_NAMESPACE")



import json
import uuid
from json import JSONDecodeError
from typing import Tuple, Dict, List

import boto3

from melange.drivers.interfaces import Queue, Topic, MessagingDriver, Message


class AWSDriver(MessagingDriver):
    def __init__(self, **kwargs):
        super().__init__()
        self.max_number_of_messages = kwargs.get("max_number_of_messages", 10)
        self.visibility_timeout = kwargs.get("visibility_timeout", 100)
        self.wait_time_seconds = kwargs.get("wait_time_seconds", 10)

    def declare_topic(self, topic_name) -> Topic:
        sns = boto3.resource("sns")
        topic = sns.create_topic(Name=topic_name)
        return topic

    def get_queue(self, queue_name) -> Queue:
        sqs_res = boto3.resource("sqs")

        return sqs_res.get_queue_by_name(QueueName=queue_name)

    def declare_queue(
        self,
        queue_name: str,
        *topics_to_bind: Topic,
        dead_letter_queue_name: str = None,
        **kwargs
    ) -> Tuple[Queue, Queue]:
        try:
            queue = self.get_queue(queue_name)
        except Exception:
            queue = self._create_queue(queue_name, content_based_deduplication="true")

        if topics_to_bind:
            statements = []
            for topic in topics_to_bind:
                statement = {
                    "Sid": "Sid{}".format(uuid.uuid4()),
                    "Effect": "Allow",
                    "Principal": "*",
                    "Resource": queue.attributes["QueueArn"],
                    "Action": "sqs:SendMessage",
                    "Condition": {"ArnEquals": {"aws:SourceArn": topic.arn}},
                }

                statements.append(statement)
                subscription = topic.subscribe(
                    Protocol="sqs",
                    Endpoint=queue.attributes[
                        "QueueArn"
                    ],  # , Attributes={"RawMessageDelivery": "true"}
                )

                if kwargs.get("filter_events"):
                    filter_policy = {"event_type": kwargs["filter_events"]}
                else:
                    filter_policy = {}

                subscription.set_attributes(
                    AttributeName="FilterPolicy",
                    AttributeValue=json.dumps(filter_policy),
                )

            policy = {
                "Version": "2012-10-17",
                "Id": "sqspolicy",
                "Statement": statements,
            }

            queue.set_attributes(Attributes={"Policy": json.dumps(policy)})

        dead_letter_queue = None
        if dead_letter_queue_name:
            try:
                dead_letter_queue = self.get_queue(dead_letter_queue_name)
            except Exception:
                dead_letter_queue = self._create_queue(
                    dead_letter_queue_name, content_based_deduplication="true"
                )

            redrive_policy = {
                "deadLetterTargetArn": dead_letter_queue.attributes["QueueArn"],
                "maxReceiveCount": "4",
            }

            queue.set_attributes(
                Attributes={"RedrivePolicy": json.dumps(redrive_policy)}
            )

        return queue, dead_letter_queue

    def _create_queue(self, queue_name: str, **kwargs) -> Queue:
        sqs_res = boto3.resource("sqs")
        fifo = queue_name.endswith(".fifo")
        attributes = {}
        if fifo:
            attributes["FifoQueue"] = "true"
            attributes["ContentBasedDeduplication"] = (
                "true" if kwargs.get("content_based_deduplication") else "false"
            )
        queue = sqs_res.create_queue(QueueName=queue_name, Attributes=attributes)
        return queue

    def retrieve_messages(self, queue: Queue, attempt_id=None) -> List[Message]:
        kwargs = dict(
            MaxNumberOfMessages=self.max_number_of_messages,
            VisibilityTimeout=self.visibility_timeout,
            WaitTimeSeconds=self.wait_time_seconds,
            MessageAttributeNames=["All"],
            AttributeNames=["All"],
        )

        if attempt_id:
            kwargs["ReceiveRequestAttemptId"] = attempt_id

        messages = queue.receive_messages(**kwargs)

        # We need to differentiate here whether the message came from SNS or SQS

        return [self._construct_message(message) for message in messages]

    def queue_publish(
        self,
        content: str,
        queue,
        event_type_name: str = None,
        message_group_id: str = None,
        message_deduplication_id: str = None,
    ):
        kwargs = dict(MessageBody=json.dumps({"Message": content}))

        if event_type_name:
            kwargs["MessageAttributes"] = {
                "event_type": {"DataType": "String", "StringValue": event_type_name}
            }

        if message_group_id:
            kwargs["MessageGroupId"] = message_group_id

        if message_deduplication_id:
            kwargs["MessageDeduplicationId"] = message_deduplication_id

        queue.send_message(**kwargs)

    def publish(
        self,
        content: str,
        topic: Topic,
        event_type_name: str,
        extra_attributes: Dict = None,
    ):
        args = dict(
            Message=content,
            MessageAttributes={
                "event_type": {"DataType": "String", "StringValue": event_type_name}
            },
        )

        if extra_attributes:
            if "subject" in extra_attributes:
                args["Subject"] = extra_attributes["subject"]

            if "message_attributes" in extra_attributes:
                args["MessageAttributes"].update(extra_attributes["message_attributes"])

            if "message_structure" in extra_attributes:
                args["MessageStructure"] = extra_attributes["message_structure"]

        response = topic.publish(**args)

        if "MessageId" not in response:
            raise ConnectionError("Could not send the event to the SNS TOPIC")

    def acknowledge(self, message: Message) -> None:
        message.metadata.delete()

    def close_connection(self) -> None:
        pass

    def delete_queue(self, queue: Queue) -> None:
        queue.delete()

    def delete_topic(self, topic: Topic) -> None:
        topic.delete()

    def _construct_message(self, message) -> Message:
        body = message.body
        manifest = ""
        try:
            message_content = json.loads(body)
            if "Message" in message_content:
                content = message_content["Message"]
                # Does the content have more attributes? If so, it is very likely that the message came from a non-raw
                # SNS redirection
                if "MessageAttributes" in message_content:
                    manifest = (
                        message_content["MessageAttributes"]
                        .get("event_type", {})
                        .get("Value")
                        or ""
                    )
            else:
                content = message_content
        except JSONDecodeError:
            content = body

        manifest = (
            manifest
            or message.message_attributes.get("event_type", {}).get("StringValue")
            or ""
        )

        return Message(message.message_id, content, message, manifest)



from redis_cache import SimpleCache, logging
from singleton import Singleton

from melange import settings


logger = logging.getLogger(__name__)


class DedupCache:
    def store(self, key, value, expire=None):
        raise NotImplementedError

    def get(self, key):
        raise NotImplementedError


class Cache(metaclass=Singleton):
    def __init__(self):
        self.cache = SimpleCache(
            expire=3600,
            host=settings.CACHE_REDIS_HOST,
            port=settings.CACHE_REDIS_PORT,
            db=settings.CACHE_REDIS_DB,
            password=settings.CACHE_REDIS_PASSWORD,
            namespace=settings.CACHE_NAMESPACE,
        )

        if not self.cache.connection:
            logger.warning(
                "Could not establish a connection with redis. Message deduplication won't work"
            )

    def store(self, key, value, expire=None):
        if not self.cache.connection:
            return

        return self.cache.store(key, value, expire)

    def get(self, key):
        if not self.cache.connection:
            return

        return self.cache.get(key)

    def __contains__(self, key):
        return self.cache.connection and key in self.cache



from melange.drivers.driver_manager import DriverManager
from melange.drivers.aws.aws_driver import AWSDriver
from melange.drivers.rabbitmq.rabbitmq_driver import RabbitMQDriver

DriverManager().add_available_drivers(aws=AWSDriver, rabbitMQ=RabbitMQDriver)



import json
from typing import Generic, Dict, Optional

from melange.helpers.typing import T


class MessageSerializer(Generic[T]):
    """
    You need to provide a way to convert a message from your sqs
    into something meaningful for your domain (e.g. into a Domain Event)
    """

    def manifest(self, data: T) -> str:
        return ""

    def deserialize(self, data: str, manifest: Optional[str] = None) -> str:
        pass

    def serialize(self, data: T) -> str:
        """
        Serializes and object to a string representation
        """
        pass


class JsonSQSSerializer(MessageSerializer[Dict]):
    def manifest(self, data: Dict):
        return "json"

    def deserialize(self, serialized_data: str, manifest: Optional[str] = None) -> str:
        data = json.loads(serialized_data)
        return data

    def serialize(self, data: Dict) -> str:
        return json.dumps(data)



import logging
from typing import Any

from funcy import lmap
from methoddispatch import singledispatch, SingleDispatch

logger = logging.getLogger(__name__)


class ExchangeListener(SingleDispatch):
    """
    The domain event serializer is used to read the event data and then generate a proper
    object that can be used by the process dispatcher
    """

    def process(self, obj: Any, **kwargs):
        self._process(obj)

    @singledispatch
    def _process(self, event):
        """Event should be an instance of DomainEvent"""
        pass

    def listens_to(self):
        accepted_events = filter(lambda t: t is not object, self._process.registry)
        return lmap(lambda ev_type: ev_type.__name__, accepted_events)

    def accepts(self, manifest: str):
        """
        Default implementation. You can override this if you want, for example,
        to accept any manifest and not only the type of classes you listen
        (useful to override in the face of subclasses)
        """
        return not self.listens_to() or manifest in self.listens_to()


listener = ExchangeListener._process.register



from singleton import Singleton

from melange.drivers.interfaces import MessagingDriver


class DriverManager(metaclass=Singleton):
    """
    This class should be used to initialize the type of messaging provider you
    want to use (Rabbit, AWS, etc)
    """

    def __init__(self):
        self._driver = None
        self._drivers = {}

    def add_available_drivers(self, **kwargs):
        self._drivers.update(kwargs)

    def use_driver(self, driver: MessagingDriver = None, driver_name=None, **kwargs):
        if driver:
            if not isinstance(driver, MessagingDriver):
                raise Exception("Invalid driver supplied")

            self._driver = driver

        elif driver_name:
            driver = self._drivers.get(driver_name)
            if not driver:
                raise Exception("Invalid driver supplied")

            self._driver = driver(**kwargs)

        else:
            raise Exception("You need to either supply a driver or a driver_name!")

    def get_driver(self):
        if not self._driver:
            raise Exception(
                "No driver is registered. Please call 'use_driver' prior to getting it"
            )

        return self._driver



import uuid

from melange.drivers.rabbitmq.rabbitmq_driver import RabbitMQDriver


class TestRabbitMQDriver:
    def setup_method(self):
        self.topic = None
        self.queue = None
        self.dead_letter_queue = None

        self.driver = RabbitMQDriver(host="localhost")

    def teardown_method(self):
        if self.topic:
            self.driver.channel.exchange_delete(exchange=self.topic)

        if self.queue:
            self.driver.channel.queue_delete(queue=self.queue)

        if self.dead_letter_queue:
            self.dead_letter_queue.delete()

    def test_create_topic(self):
        topic_name = self._get_topic_name()
        self.topic = self.driver.declare_topic(topic_name)
        assert self.topic == topic_name

    def test_create_normal_queue(self):
        queue_name = self._get_queue_name()
        self.queue, _ = self.driver.declare_queue(queue_name)
        assert self.queue == queue_name

    def test_create_queue_and_bind_to_topic(self):
        self.topic = self.driver.declare_topic(self._get_topic_name())
        queue_name = self._get_queue_name()
        self.queue, _ = self.driver.declare_queue(queue_name, self.topic)

        assert self.queue == queue_name

    # def test_create_queue_with_dead_letter_queue(self):
    #     self.topic = self.driver.declare_topic(self._get_topic_name())
    #     self.queue, self.dead_letter_queue = self.driver.declare_queue(self._get_queue_name(),
    #                                                      self.topic,
    #                                                      dead_letter_queue_name=self._get_queue_name())
    #
    #     assert self.queue.url
    #
    #     attrs = self.queue.attributes
    #
    #     policy = json.loads(attrs['Policy'])
    #     policy_topic_arn = policy['Statement'][0]['Condition']['ArnEquals']['aws:SourceArn']
    #     assert policy_topic_arn == self.topic.arn
    #
    #     redrive_policy = json.loads(attrs['RedrivePolicy'])
    #     queue_arn = self.dead_letter_queue.attributes['QueueArn']
    #
    #     assert redrive_policy['deadLetterTargetArn'] == queue_arn
    #     assert redrive_policy['maxReceiveCount'] == 4

    def _get_queue_name(self):
        return "test_queue_{}".format(uuid.uuid4())

    def _get_topic_name(self):
        return "test_queue_{}".format(uuid.uuid4())



import logging
from typing import List

from melange.drivers.driver_manager import DriverManager
from melange.event_serializer import MessageSerializer
from melange.exchange_listener import ExchangeListener
from melange.infrastructure.cache import DedupCache, Cache
from melange.drivers.interfaces import MessagingDriver, Message
from melange.utils import get_fully_qualified_name

logger = logging.getLogger(__name__)


class ExchangeMessageYielder:
    def __init__(
        self,
        event_queue_name,
        message_serializer: MessageSerializer,
        *topics_to_subscribe,
        dead_letter_queue_name=None,
        driver=None,
        **kwargs
    ):
        self._driver = driver or DriverManager().get_driver()
        self.message_serializer = message_serializer
        self._event_queue_name = event_queue_name
        self._topics_to_subscribe = topics_to_subscribe
        self._dead_letter_queue_name = dead_letter_queue_name

        self._topics = [
            self._driver.declare_topic(t) for t in self._topics_to_subscribe
        ]
        self._event_queue, self._dead_letter_queue = self._driver.declare_queue(
            self._event_queue_name,
            *self._topics,
            dead_letter_queue_name=self._dead_letter_queue_name,
            filter_events=kwargs.get("filter_events")
        )

    def get_messages(self):
        event_queue = self._driver.get_queue(self._event_queue_name)

        while True:
            messages = self._driver.retrieve_messages(event_queue)

            for message in messages:
                if "event_type_name" in message.content:
                    yield self.message_serializer.deserialize(message.content)

    def acknowledge(self, message):
        self._driver.acknowledge(message)


class ExchangeMessageConsumer:
    def __init__(
        self,
        event_queue_name: str,
        message_serializer: MessageSerializer,
        *topic_names_to_subscribe: str,
        dead_letter_queue_name: str = None,
        cache: DedupCache = None,
        driver: MessagingDriver = None,
        **kwargs
    ):
        self._exchange_listeners: List[ExchangeListener] = []
        self.message_serializer = message_serializer
        self._driver = driver or DriverManager().get_driver()
        self._event_queue_name = event_queue_name
        self._topics_to_subscribe = topic_names_to_subscribe
        self._dead_letter_queue_name = dead_letter_queue_name
        self.cache = cache or Cache()

        self._topics = [
            self._driver.declare_topic(t) for t in self._topics_to_subscribe
        ]
        self._event_queue, self._dead_letter_queue = self._driver.declare_queue(
            self._event_queue_name,
            *self._topics,
            dead_letter_queue_name=self._dead_letter_queue_name,
            filter_events=kwargs.get("filter_events")
        )

    def subscribe(self, exchange_listener: ExchangeListener) -> None:
        if not isinstance(exchange_listener, ExchangeListener):
            return None

        if exchange_listener not in self._exchange_listeners:
            self._exchange_listeners.append(exchange_listener)

    def unsubscribe(self, exchange_listener: ExchangeListener) -> None:
        if exchange_listener in self._exchange_listeners:
            self._exchange_listeners.remove(exchange_listener)

    def consume_event(self):
        event_queue = self._driver.get_queue(self._event_queue_name)

        messages = self._driver.retrieve_messages(event_queue)

        for message in messages:
            try:
                self._process_message(message)
            except Exception as e:
                logger.exception(e)

    def _get_subscribers(self, manifest: str):
        return [
            listener
            for listener in self._exchange_listeners
            if listener.accepts(manifest)
        ]

    def _process_message(self, message: Message):
        manifest = message.get_message_manifest()
        message_data = self.message_serializer.deserialize(
            message.content, manifest=manifest
        )
        subscribers = self._get_subscribers(manifest)

        successful = 0
        for subscr in subscribers:
            try:
                # Store into the cache
                message_listener_key = (
                    get_fully_qualified_name(subscr) + "." + message.message_id
                )

                if message_listener_key in self.cache:
                    logger.info("detected a duplicated message, ignoring")
                else:
                    subscr.process(message_data, message_id=message.message_id)
                    successful += 1
                    self.cache.store(message_listener_key, message_listener_key)
            except Exception as e:
                logger.exception(e)

        if successful == len(subscribers):
            self._driver.acknowledge(message)



import json
import time
import uuid

import pytest
from hamcrest import *

from melange import AWSDriver


@pytest.fixture
def aws_driver():
    return AWSDriver(wait_time_seconds=1, visibility_timeout=10)


@pytest.fixture
def topic(aws_driver):
    topic = aws_driver.declare_topic("dev-topic-{}".format(uuid.uuid4()))
    yield topic
    aws_driver.delete_topic(topic)


def test_create_fifo_queue(aws_driver, request):
    def delete_queue():
        aws_driver.delete_queue(queue)

    request.addfinalizer(delete_queue)

    queue_name = "test-queue-{}.fifo".format(uuid.uuid4())
    aws_driver.declare_queue(queue_name)
    queue = aws_driver.get_queue(queue_name)

    assert_that(
        queue.attributes,
        has_entries(FifoQueue="true", ContentBasedDeduplication="true"),
    )


def test_create_fifo_queue_with_dlq(aws_driver, request):
    def delete_queue():
        aws_driver.delete_queue(queue)
        aws_driver.delete_queue(dlq)

    request.addfinalizer(delete_queue)

    queue_name = "test-queue-{}.fifo".format(uuid.uuid4())
    dlq_queue_name = "test-queue-{}-dlq.fifo".format(uuid.uuid4())
    queue, dlq = aws_driver.declare_queue(
        queue_name, dead_letter_queue_name=dlq_queue_name
    )

    assert_that(
        queue.attributes,
        has_entries(FifoQueue="true", ContentBasedDeduplication="true"),
    )
    assert_that(
        dlq.attributes, has_entries(FifoQueue="true", ContentBasedDeduplication="false")
    )


def test_create_a_non_fifo_queue(aws_driver, request):
    def delete_queue():
        aws_driver.delete_queue(queue)

    request.addfinalizer(delete_queue)

    queue_name = "test-queue-{}".format(uuid.uuid4())
    aws_driver.declare_queue(queue_name)
    queue = aws_driver.get_queue(queue_name)

    assert_that(
        queue.attributes,
        not_(has_entries(FifoQueue="true", ContentBasedDeduplication="true")),
    )


def test_create_normal_queue_with_dlq(aws_driver, request):
    def delete_queue():
        aws_driver.delete_queue(queue)
        aws_driver.delete_queue(dlq)

    request.addfinalizer(delete_queue)

    queue_name = "test-queue-{}".format(uuid.uuid4())
    dlq_queue_name = "test-queue-{}-dlq".format(uuid.uuid4())
    queue, dlq = aws_driver.declare_queue(
        queue_name, dead_letter_queue_name=dlq_queue_name
    )

    assert_that(
        queue.attributes,
        not_(has_entries(FifoQueue="true", ContentBasedDeduplication="true")),
    )
    assert_that(
        dlq.attributes,
        not_(has_entries(FifoQueue="true", ContentBasedDeduplication="false")),
    )


def test_send_messages_through_a_fifo_queue_and_make_sure_they_are_always_ordered(
    aws_driver, request
):
    def delete_queue():
        aws_driver.delete_queue(queue)

    request.addfinalizer(delete_queue)

    queue_name = "test-queue-{}.fifo".format(uuid.uuid4())
    queue, _ = aws_driver.declare_queue(queue_name)

    message_group_id = "my-app"

    for i in range(100):
        message_dedup_id = str(uuid.uuid4())
        aws_driver.queue_publish(
            f"my-message-{i}",
            queue,
            "my-event-type",
            message_group_id,
            message_dedup_id,
        )

    received_messages = []
    retrieved_messages = aws_driver.retrieve_messages(queue)

    while retrieved_messages:
        for m in retrieved_messages:
            received_messages.append(m.content)
            aws_driver.acknowledge(m)

        retrieved_messages = aws_driver.retrieve_messages(queue)

    expected_array = [f"my-message-{i}" for i in range(100)]
    assert_that(received_messages, contains(*expected_array))


def test_not_acknowledging_any_of_the_messages_on_a_fifo_queue_will_delay_the_delivery_of_the_rest(
    aws_driver, request
):
    def delete_queue():
        aws_driver.delete_queue(queue)

    request.addfinalizer(delete_queue)

    queue_name = "test-queue-{}.fifo".format(uuid.uuid4())
    queue, _ = aws_driver.declare_queue(queue_name)

    message_group_id = "my-app"

    for i in range(20):
        message_dedup_id = str(uuid.uuid4())
        aws_driver.queue_publish(
            f"my-message-{i}",
            queue,
            "my-event-type",
            message_group_id,
            message_dedup_id,
        )

    received_messages = []
    retrieved_messages = aws_driver.retrieve_messages(queue)

    num_acks = 0
    while num_acks < 20:
        for i, m in enumerate(retrieved_messages):
            if i == 5:
                break
            else:
                received_messages.append(m.content)
                aws_driver.acknowledge(m)
                num_acks += 1
        retrieved_messages = aws_driver.retrieve_messages(queue)

    expected_array = [f"my-message-{i}" for i in range(20)]
    assert_that(received_messages, contains(*expected_array))


def test_send_messages_through_a_non_fifo_queue_does_not_guarantee_order(
    aws_driver, request
):
    def delete_queue():
        aws_driver.delete_queue(queue)

    request.addfinalizer(delete_queue)

    queue_name = "test-queue-{}".format(uuid.uuid4())
    queue, _ = aws_driver.declare_queue(queue_name)

    for i in range(100):
        aws_driver.queue_publish(f"my-message-{i}", queue, "my-event-type")

    received_messages = []
    retrieved_messages = aws_driver.retrieve_messages(queue)

    while retrieved_messages:
        for m in retrieved_messages:
            received_messages.append(m.content)
            aws_driver.acknowledge(m)

        retrieved_messages = aws_driver.retrieve_messages(queue)

    expected_array = [f"my-message-{i}" for i in range(100)]
    assert_that(received_messages, not_(contains(*expected_array)))
    assert_that(received_messages, contains_inanyorder(*expected_array))


def test_aws_driver_declare_a_queue_for_a_topic_filtering_the_events_it_sends_to_certain_queues(
    aws_driver, topic
):
    try:
        queue_1, _ = aws_driver.declare_queue(
            "dev-queue-{}".format(uuid.uuid4()), topic, filter_events=["MyBananaEvent"]
        )
        queue_2, _ = aws_driver.declare_queue(
            "dev-queue-{}".format(uuid.uuid4()),
            topic,
            filter_events=["MyNonBananaEvent"],
        )
        queue_3, _ = aws_driver.declare_queue(
            "dev-queue-{}".format(uuid.uuid4()), topic
        )

        aws_driver.publish(
            json.dumps({"event_type_name": "MyBananaEvent", "value": 3}),
            topic,
            event_type_name="MyBananaEvent",
        )
        aws_driver.publish(
            json.dumps({"event_type_name": "MyNonBananaEvent", "value": 5}),
            topic,
            event_type_name="MyNonBananaEvent",
        )
        time.sleep(2)

        _assert_messages(aws_driver, queue_3, ["MyNonBananaEvent", "MyBananaEvent"], 2)
        _assert_messages(aws_driver, queue_1, ["MyBananaEvent"], 1)
        _assert_messages(aws_driver, queue_2, ["MyNonBananaEvent"], 1)

    finally:
        aws_driver.delete_queue(queue_1)
        aws_driver.delete_queue(queue_2)
        aws_driver.delete_queue(queue_3)


def _assert_messages(
    aws_driver, queue, expected_types_contained, expected_num_messages
):
    messages = []
    while True:
        retrieved_messages = aws_driver.retrieve_messages(queue)
        messages += retrieved_messages
        if len(messages) > expected_num_messages:
            raise Exception("The test has failed")
        if not retrieved_messages:
            break

    assert expected_num_messages == len(messages)

    received_event_types = [message.content["event_type_name"] for message in messages]
    for type in expected_types_contained:
        assert type in received_event_types


class TestAWSDriver:
    def setup_method(self):
        self.topic = None
        self.topic_1 = None
        self.topic_2 = None
        self.queue = None
        self.dead_letter_queue = None

        self.driver = AWSDriver()

    def teardown_method(self):
        if self.topic:
            self.topic.delete()

        if self.topic_1:
            self.topic_1.delete()

        if self.topic_2:
            self.topic_2.delete()

        if self.queue:
            self.queue.delete()

        if self.dead_letter_queue:
            self.dead_letter_queue.delete()

    def test_create_topic(self):
        self.topic = self.driver.declare_topic(self._get_topic_name())
        assert self.topic.arn

    def test_create_normal_queue(self):
        self.queue, _ = self.driver.declare_queue(self._get_queue_name())
        assert self.queue.url
        assert "Policy" not in self.queue.attributes

    def test_create_queue_and_bind_to_topic(self):
        self.topic = self.driver.declare_topic(self._get_topic_name())
        self.queue, _ = self.driver.declare_queue(self._get_queue_name(), self.topic)

        assert self.queue.url

        policy = json.loads(self.queue.attributes["Policy"])
        policy_topic_arn = policy["Statement"][0]["Condition"]["ArnEquals"][
            "aws:SourceArn"
        ]
        assert policy_topic_arn == self.topic.arn

    def test_create_queue_and_bind_several_topics(self):
        self.topic_1 = self.driver.declare_topic(self._get_topic_name())
        self.topic_2 = self.driver.declare_topic(self._get_topic_name())
        self.queue, _ = self.driver.declare_queue(
            self._get_queue_name(), self.topic_1, self.topic_2
        )

        assert self.queue.url

        policy = json.loads(self.queue.attributes["Policy"])
        policy_topic_arn = policy["Statement"][0]["Condition"]["ArnEquals"][
            "aws:SourceArn"
        ]
        assert policy_topic_arn == self.topic_1.arn

        policy_topic_arn = policy["Statement"][1]["Condition"]["ArnEquals"][
            "aws:SourceArn"
        ]
        assert policy_topic_arn == self.topic_2.arn

    def test_create_queue_with_dead_letter_queue(self):
        self.topic = self.driver.declare_topic(self._get_topic_name())
        self.queue, self.dead_letter_queue = self.driver.declare_queue(
            self._get_queue_name(),
            self.topic,
            dead_letter_queue_name=self._get_queue_name(),
        )

        assert self.queue.url

        attrs = self.queue.attributes

        policy = json.loads(attrs["Policy"])
        policy_topic_arn = policy["Statement"][0]["Condition"]["ArnEquals"][
            "aws:SourceArn"
        ]
        assert policy_topic_arn == self.topic.arn

        redrive_policy = json.loads(attrs["RedrivePolicy"])
        queue_arn = self.dead_letter_queue.attributes["QueueArn"]

        assert redrive_policy["deadLetterTargetArn"] == queue_arn
        assert redrive_policy["maxReceiveCount"] == 4

    def _get_queue_name(self):
        return "test_queue_{}".format(uuid.uuid4())

    def _get_topic_name(self):
        return "test_queue_{}".format(uuid.uuid4())



from melange.drivers.driver_manager import DriverManager


def configure_exchange(driver_name, **kwargs):
    DriverManager().use_driver(
        driver_name=driver_name,
        wait_time_seconds=kwargs.get("wait_time_seconds", 10),
        visibility_timeout=kwargs.get("visibility_timeout", 10),
    )



from uuid import uuid4

from melange.drivers.interfaces import MessagingDriver, Message


class DumbQueue:
    def __init__(self, name):
        self.name = name
        self.attributes = {}


class InMemoryMessagingDriver(MessagingDriver):
    def __init__(self, callback=None, only_queues=None):
        super().__init__()
        self._messages = []
        self.callback = callback
        self._only_queues = only_queues or []

    def set_callback(self, callback):
        self.callback = callback

    def declare_topic(self, topic_name):
        return None

    def get_queue(self, queue_name):
        return DumbQueue(queue_name)

    def declare_queue(
        self, queue_name, *topics_to_bind, dead_letter_queue_name=None, **kwargs
    ):
        return DumbQueue(queue_name), None

    def retrieve_messages(self, queue, attempt_id=None):
        messages = self._messages
        self._messages = []
        return messages

    def publish(self, content, topic, event_type_name, extra_attributes=None):
        if not self._only_queues or topic.name in self._only_queues:
            message = Message(str(uuid4()), content, None, manifest=event_type_name)
            self._messages.append(message)

            if self.callback:
                self.callback(message)

    def queue_publish(
        self,
        content,
        queue,
        event_type_name,
        message_group_id=None,
        message_deduplication_id=None,
    ):
        self.publish(content, queue, event_type_name)

    def acknowledge(self, message):
        return None

    def delete_queue(self, queue):
        return None

    def delete_topic(self, topic):
        return None



from unittest.mock import MagicMock

from melange.messaging import MessagingDriver, DriverManager, ExchangeMessagePublisher


class TestMessagePublisher:
    def test_publish_a_message(self):
        a_topic = "a_topic"
        event = {"some_content": "12345"}

        driver = MagicMock(spec=MessagingDriver)
        DriverManager().use_driver(driver=driver)

        message_publisher = ExchangeMessagePublisher(a_topic)
        success = message_publisher.publish(event, manifest="some_event")

        assert success
        driver.publish.assert_called()
